@@ -1,33 +1,48 @@
-(* File: conv.mli
-
-    Copyright (C) 2005-
-
-      Jane Street Holding, LLC
-      Author: Markus Mottl
-      email: mmottl\@janestcapital.com
-      WWW: http://www.janestcapital.com/ocaml
-
-   This library is free software; you can redistribute it and/or
-   modify it under the terms of the GNU Lesser General Public
-   License as published by the Free Software Foundation; either
-   version 2 of the License, or (at your option) any later version.
-
-   This library is distributed in the hope that it will be useful,
-   but WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   Lesser General Public License for more details.
-
-   You should have received a copy of the GNU Lesser General Public
-   License along with this library; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-*)
+(******************************************************************************
+ *                             Sexplib                                        *
+ *                                                                            *
+ * Copyright (C) 2005- Jane Street Holding, LLC                               *
+ *    Contact: opensource@janestreet.com                                      *
+ *    WWW: http://www.janestreet.com/ocaml                                    *
+ *    Author: Markus Mottl                                                    *
+ *                                                                            *
+ * This library is free software; you can redistribute it and/or              *
+ * modify it under the terms of the GNU Lesser General Public                 *
+ * License as published by the Free Software Foundation; either               *
+ * version 2 of the License, or (at your option) any later version.           *
+ *                                                                            *
+ * This library is distributed in the hope that it will be useful,            *
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
+ * Lesser General Public License for more details.                            *
+ *                                                                            *
+ * You should have received a copy of the GNU Lesser General Public           *
+ * License along with this library; if not, write to the Free Software        *
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
+ *                                                                            *
+ ******************************************************************************)
 
 (** Conv: Utility Module for S-expression Conversions *)
 
 open Bigarray
 
-(** Dummy definition for "optional" options *)
+(** Dummy definitions for "optional" options, lists, and for opaque types *)
+type sexp_bool = bool
 type 'a sexp_option = 'a option
+type 'a sexp_list = 'a list
+type 'a sexp_array = 'a array
+type 'a sexp_opaque = 'a
+
+
+(** {6 Type aliases} *)
+
+type bigstring = Sexp.bigstring
+type float32_vec = (float, float32_elt, fortran_layout) Array1.t
+type float64_vec = (float, float64_elt, fortran_layout) Array1.t
+type vec = float64_vec
+type float32_mat = (float, float32_elt, fortran_layout) Array2.t
+type float64_mat = (float, float64_elt, fortran_layout) Array2.t
+type mat = float64_mat
 
 
 (** {6 Conversion of OCaml-values to S-expressions} *)
@@ -47,15 +62,23 @@ val write_old_option_format : bool ref
     Initially set to [true].
 *)
 
+
 val read_old_option_format : bool ref
 (** [read_old_option_format] reference for the default option format
-    used to read option values.  [Of_sexp_error] will be raised with
-    old-style option values if this reference is set to [false].
-    Reading new-style option values is always supported.
+    used to read option values.  [Of_sexp_error] will be raised
+    with old-style option values if this reference is set to [false].
+    Reading new-style option values is always supported.  Using a global
+    reference instead of changing the converter calling conventions is
+    the only way to avoid breaking old code with the standard macros.
 
     Initially set to [true].
 *)
 
+(** We re-export a tail recursive map function, because some modules
+    override the standard library functions (e.g. [StdLabels]) which
+    wrecks havoc with the camlp4 extension. *)
+val list_map : ('a -> 'b) -> 'a list -> 'b list
+
 val sexp_of_unit : unit -> Sexp.t
 (** [sexp_of_unit ()] converts a value of type [unit] to an S-expression. *)
 
@@ -150,49 +173,41 @@ val sexp_of_hashtbl :
     to convert the hashtable keys of type ['a], and [conv_value] to
     convert hashtable values of type ['b] to S-expressions. *)
 
-val sexp_of_float32_vec :
-  (float, float32_elt, fortran_layout) Array1.t -> Sexp.t
+val sexp_of_bigstring : bigstring -> Sexp.t
+(** [sexp_of_bigstring bstr] converts a bigstring (character bigarray
+    in C-layout) to an S-expression. *)
+
+val sexp_of_float32_vec : float32_vec -> Sexp.t
 (** [sexp_of_float32_vec vec] converts the one-dimensional bigarray
     [vec] of 32-bit floats in Fortran-layout to an S-expression. *)
 
-val sexp_of_float64_vec :
-  (float, float64_elt, fortran_layout) Array1.t -> Sexp.t
+val sexp_of_float64_vec : float64_vec -> Sexp.t
 (** [sexp_of_float64_vec vec] converts the one-dimensional bigarray
     [vec] of 64-bit floats in Fortran-layout to an S-expression. *)
 
-val sexp_of_vec : (float, float64_elt, fortran_layout) Array1.t -> Sexp.t
+val sexp_of_vec : vec -> Sexp.t
 (** [sexp_of_vec vec] same as {!Conv.sexp_of_float64_vec}. *)
 
-val sexp_of_float32_mat :
-  (float, float32_elt, fortran_layout) Array2.t -> Sexp.t
+val sexp_of_float32_mat : float32_mat -> Sexp.t
 (** [sexp_of_float32_mat mat] converts the two-dimensional bigarray
     [mat] of 32-bit floats in Fortran-layout to an S-expression. *)
 
-val sexp_of_float64_mat :
-  (float, float64_elt, fortran_layout) Array2.t -> Sexp.t
+val sexp_of_float64_mat : float64_mat -> Sexp.t
 (** [sexp_of_float64_mat mat] converts the two-dimensional bigarray
     [mat] of 64-bit floats in Fortran-layout to an S-expression. *)
 
-val sexp_of_mat : (float, float64_elt, fortran_layout) Array2.t -> Sexp.t
+val sexp_of_mat : mat -> Sexp.t
 (** [sexp_of_mat mat] same as {!Conv.sexp_of_float64_mat}. *)
 
-val sexp_of_abstr : 'a -> Sexp.t
-(** [sexp_of_abstr x] converts the value [x] of abstract type to an
-    S-expression. *)
+val sexp_of_opaque : 'a -> Sexp.t
+(** [sexp_of_opaque x] converts the value [x] of opaque type to an
+    S-expression.  This means the user need not provide converters,
+    but the result cannot be interpreted. *)
 
-val sexp_of_fun : ('a -> 'b) -> Sexp.t
-(** [sexp_of_fun f] converts the value [f] of function type to an
-    S-expression. *)
-
-(** Type used for declaring existing types as opaque to sexp converters,
-    i.e. even if there is a defined converter for ['a], it will not
-    be called.  This is useful to e.g. suppress printing of extremely
-    large sub-structures for purely informational, human-readable output.
-*)
-type 'a opaque = 'a
-
-val sexp_of_opaque : ('a -> Sexp.t) -> 'a opaque -> Sexp.t
-(** [sexp_of_opaque _ _] converts opaque OCaml-values to S-expressions. *)
+val sexp_of_fun : _ -> Sexp.t
+(** [sexp_of_fun f] converts the value [f] of function type to a
+    dummy S-expression.  Functions cannot be serialized as S-expressions,
+    but at least a placeholder can be generated for pretty-printing. *)
 
 val string_of__of__sexp_of : ('a -> Sexp.t) -> 'a -> string
 (** [string_of__of__sexp_of conv x] converts the OCaml-value [x] to
@@ -202,17 +217,19 @@ val string_of__of__sexp_of : ('a -> Sexp.t) -> 'a -> string
 
 (** {6 Conversion of S-expressions to OCaml-values} *)
 
-exception Of_sexp_error of string * Sexp.t
-(** [Of_sexp_error (reason, sexp)] the exception raised when an
-    S-expression could not be successfully converted to an OCaml-value. *)
+exception Of_sexp_error of exn * Sexp.t
+(** [Of_sexp_error (exn, sexp)] the exception raised when an S-expression
+    could not be successfully converted to an OCaml-value. *)
 
 val record_check_extra_fields : bool ref
 (** [record_check_extra_fields] checks for extra (= unknown) fields
     in record S-expressions. *)
 
 val of_sexp_error : string -> Sexp.t -> 'a
-(** [of_sexp_error reason sexp] @raise the exception [Of_sexp_error
-    (reason, sexp)]. *)
+(** [of_sexp_error reason sexp] @raise Of_sexp_error (Failure reason, sexp). *)
+
+val of_sexp_error_exn : exn -> Sexp.t -> 'a
+(** [of_sexp_error exc sexp] @raise Of_sexp_error (exc, sexp). *)
 
 val unit_of_sexp : Sexp.t -> unit
 (** [unit_of_sexp sexp] converts S-expression [sexp] to a value of type
@@ -312,37 +329,103 @@ val hashtbl_of_sexp :
     key of type ['a], and function [conv_value], which converts an
     S-expression to hashtable value of type ['b]. *)
 
-val float32_vec_of_sexp :
-  Sexp.t -> (float, float32_elt, fortran_layout) Array1.t
+val float32_vec_of_sexp : Sexp.t -> float32_vec
 (** [float32_vec_of_sexp sexp] converts S-expression [sexp] to a
     one-dimensional bigarray of 32-bit floats in Fortran-layout. *)
 
-val float64_vec_of_sexp :
-  Sexp.t -> (float, float64_elt, fortran_layout) Array1.t
+val float64_vec_of_sexp : Sexp.t -> float64_vec
 (** [float64_vec_of_sexp sexp] converts S-expression [sexp] to a
     one-dimensional bigarray of 64-bit floats in Fortran-layout. *)
 
-val vec_of_sexp : Sexp.t -> (float, float64_elt, fortran_layout) Array1.t
+val vec_of_sexp : Sexp.t -> vec
 (** [vec_of_sexp sexp] same as {!float64_vec_of_sexp}. *)
 
-val float32_mat_of_sexp :
-  Sexp.t -> (float, float32_elt, fortran_layout) Array2.t
+val float32_mat_of_sexp : Sexp.t -> float32_mat
 (** [float32_mat_of_sexp sexp] converts S-expression [sexp] to a
     two-dimensional bigarray of 32-bit floats in Fortran-layout. *)
 
-val float64_mat_of_sexp :
-  Sexp.t -> (float, float64_elt, fortran_layout) Array2.t
+val float64_mat_of_sexp : Sexp.t -> float64_mat
 (** [float64_mat_of_sexp sexp] converts S-expression [sexp] to a
     two-dimensional bigarray of 64-bit floats in Fortran-layout. *)
 
-val mat_of_sexp : Sexp.t -> (float, float64_elt, fortran_layout) Array2.t
+val mat_of_sexp : Sexp.t -> mat
 (** [mat_of_sexp sexp] same as {!Conv.float64_mat_of_sexp}. *)
 
+val opaque_of_sexp : Sexp.t -> 'a
+(** [opaque_of_sexp sexp] @raise Of_sexp_error when attempting to
+    convert an S-expression to an opaque value. *)
+
 val fun_of_sexp : Sexp.t -> ('a -> 'b)
-(** [fun_of_sexp sexp] @raise a conversion error when attempting to
+(** [fun_of_sexp sexp] @raise Of_sexp_error when attempting to
     convert an S-expression to a function. *)
 
 val of_string__of__of_sexp : (Sexp.t -> 'a) -> string -> 'a
 (** [of_string__of__of_sexp conv str] converts the S-expression [str]
     represented as a string to an OCaml-value by using conversion function
     [conv]. *)
+
+
+(** Exception converters *)
+
+val sexp_of_exn : exn -> Sexp.t
+(** [sexp_of_exn exc] converts exception [exc] to an S-expression.
+    If no suitable converter is found, the standard converter in
+    [Printexc] will be used to generate an atomic S-expression. *)
+
+val sexp_of_exn_opt : exn -> Sexp.t option
+(** [sexp_of_exn_opt exc] converts exception [exc] to [Some sexp].
+    If no suitable converter is found, [None] is returned instead. *)
+
+module Exn_converter : sig
+  type t  (** Type of handles for exception S-expression converters *)
+
+  val set_max_exn_tags : int -> unit
+  (** [set_max_exn_tags n] sets the maximum number of converters for exceptions
+      with the same template to [n].  If already existing handlers exceed
+      this number, they will remain at their current number until this number
+      is reduced due to garbage collection.  New handlers will not be added
+      until [n] will not be exceeded. *)
+
+  val get_max_exn_tags : unit -> int
+  (** [set_max_exn_tags ()] return the maximum number of converters for
+      exceptions with the same template. *)
+
+  val add_auto : ?finalise : bool -> exn -> (exn -> Sexp.t) -> unit
+  (** [add_auto ?finalise templ sexp_of_exn] registers exception S-expression
+      converter [sexp_of_exn] for exceptions having same constructor as
+      template [templ], unless the number of stored handlers for the given
+      template exceeds [get_max_exn_tags ()], in which case the handler will
+      never be called.  When [sexp_of_exn] is called, the passed exception
+      is guaranteed to match the template.
+
+      NOTE: if the exception belongs to a transient module, e.g. local modules
+      (including functor instantiations), first-class modules, etc., a manually
+      written [sexp_of_exn] must use [Obj.magic] internally to avoid matching
+      or creating the exception, otherwise the handler can never be reclaimed
+      once the exception ceases to exist.  If [finalise] is [true], then
+      the exception will be automatically registered for removal with the GC
+      (default).  Finalisation will not work with exceptions that have been
+      allocated outside the heap, which is the case for some standard ones
+      e.g. [Sys_error].
+
+      NOTE: Use with great caution, this function is primarily intended for
+      automated use!  If unsure, use [add_slow] instead.
+
+      @param finalise default = [true]
+  *)
+
+  val add_slow : (exn -> Sexp.t option) -> t
+  (** [add_slow sexp_of_exn] registers exception S-expression converter
+      [sexp_of_exn] and returns a handle.  Exception converters registered this
+      way are much slower than with [add], but this function does not require
+      an exception template.  NOTE: if you call this function explicitly,
+      or the "sexp"-macro for exceptions from within local modules, you will
+      eventually have to unregister it manually with {!del}, otherwise there
+      is a space leak! *)
+
+  val del_slow : t -> unit
+  (** [del_slow handle] unregisters exception S-expression converter with
+      handle [handle].  In multi-threaded contexts it is not guaranteed
+      that the unregistered converter will not be called after this function
+      returns. *)
+end
