@@ -1,3 +1,27 @@
+(******************************************************************************
+ *                             Sexplib                                        *
+ *                                                                            *
+ * Copyright (C) 2005- Jane Street Holding, LLC                               *
+ *    Contact: opensource@janestreet.com                                      *
+ *    WWW: http://www.janestreet.com/ocaml                                    *
+ *    Author: Markus Mottl                                                    *
+ *                                                                            *
+ * This library is free software; you can redistribute it and/or              *
+ * modify it under the terms of the GNU Lesser General Public                 *
+ * License as published by the Free Software Foundation; either               *
+ * version 2 of the License, or (at your option) any later version.           *
+ *                                                                            *
+ * This library is distributed in the hope that it will be useful,            *
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
+ * Lesser General Public License for more details.                            *
+ *                                                                            *
+ * You should have received a copy of the GNU Lesser General Public           *
+ * License along with this library; if not, write to the Free Software        *
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
+ *                                                                            *
+ ******************************************************************************)
+
 (* Sexp: Module for handling S-expressions (I/O, etc.) *)
 
 open Format
@@ -17,16 +41,13 @@ let default_indent = ref 1
 
 (* Escaping of strings used as atoms in S-expressions *)
 
+let is_special_char c =
+  c <= ' ' || c = '"' || c = '(' || c = ')' || c = ';' || c = '\\'
+
 let must_escape str =
   let len = String.length str in
   len = 0 ||
-    let rec loop ix =
-      match str.[ix] with
-      | '"' | '(' | ')' | ';' | '\\' -> true
-      | '|' -> ix > 0 && let next = ix - 1 in str.[next] = '#' || loop next
-      | '#' -> ix > 0 && let next = ix - 1 in str.[next] = '|' || loop next
-      | c -> c <= ' ' || ix > 0 && loop (ix - 1)
-    in
+    let rec loop ix = is_special_char str.[ix] || ix > 0 && loop (ix - 1) in
     loop (len - 1)
 
 let maybe_esc_str str =
@@ -126,43 +147,10 @@ let to_buffer_mach ~buf sexp =
 
 let to_buffer = to_buffer_mach
 
-let to_buffer_gen ~buf ~add_char ~add_string sexp =
-  let rec loop may_need_space = function
-    | Atom str ->
-        let str' = maybe_esc_str str in
-        let new_may_need_space = str' == str in
-        if may_need_space && new_may_need_space then add_char buf ' ';
-        add_string buf str';
-        new_may_need_space
-    | List (h :: t) ->
-        add_char buf '(';
-        let may_need_space = loop false h in
-        loop_rest may_need_space t;
-        false
-    | List [] -> add_string buf "()"; false
-  and loop_rest may_need_space = function
-    | h :: t ->
-        let may_need_space = loop may_need_space h in
-        loop_rest may_need_space t
-    | [] -> add_char buf ')' in
-  ignore (loop false sexp)
-
 
 (* Output of S-expressions to I/O-channels *)
 
-(* The maximum size of a thing on the minor heap is 256 words.
-   Previously, this size of the returned buffer here was 4096 bytes, which
-   caused the Buffer to be allocated on the *major* heap every time.
-
-   According to a simple benchmark by Ron, we can improve performance for
-   small s-expressions by a factor of ~4 if we only allocate 1024 bytes
-   (128 words + some small overhead) worth of buffer initially.  And one
-   can argue that if it's free to allocate strings smaller than 256 words,
-   large s-expressions requiring larger expensive buffers won't notice
-   the extra two doublings from 1024 bytes to 2048 and 4096. And especially
-   performance-sensitive applications to always pass in a larger buffer to
-   use. *)
-let buffer () = Buffer.create 1024
+let buffer () = Buffer.create 4096
 
 let with_new_buffer oc f =
   let buf = buffer () in
@@ -183,52 +171,31 @@ let output = output_mach
 
 (* Output of S-expressions to file *)
 
-(* The temp file functions in the OCaml Filename module do not support
-   permissions.  But opening a file with given permissions is different
-   from opening it and chmoding it to these permissions, because the umask
-   is taken in account.  Under Unix there's no easy way to get the umask in
-   a thread-safe way. *)
-module Tmp_file = struct
-  let prng = ref None
-
-  let temp_file_name prefix suffix =
-    let rand_state = match !prng with
-      | Some v -> v
-      | None ->
-          let ret = Random.State.make_self_init () in
-          prng := Some ret;
-          ret
-    in
-    let rnd = (Random.State.bits rand_state) land 0xFFFFFF in
-    Printf.sprintf "%s%06x%s" prefix rnd suffix
-
-  (* Keep the permissions loose. Sexps are usually shared and rarely private*)
-  let open_temp_file ?(perm = 0o666) prefix suffix =
-    let rec try_name counter =
-      let name = temp_file_name prefix suffix in
-      try
-        let oc =
-          open_out_gen [Open_wronly; Open_creat; Open_excl; Open_text] perm name
-        in
-        name, oc
-      with Sys_error _ as e ->
-        if counter >= 1000 then raise e else try_name (counter + 1)
-    in
-    try_name 0
-end
-
 let save_of_output ?perm output_function file sexp =
-  let tmp_name, oc = Tmp_file.open_temp_file ?perm file "tmp" in
-  begin
-    try
-      output_function oc sexp;
-      close_out oc;
-    with e ->
+  let tmp_name, oc =
+    let temp_dir =
+      if Filename.is_relative file then "."
+      else Filename.dir_sep
+    in
+    Filename.open_temp_file ~temp_dir file "tmp"
+  in
+  try
+    output_function oc sexp;
+    close_out oc;
+    let perm =
+      match perm with
+      | Some perm -> perm
+      | None ->
+          let umask = Unix.umask 0 in
+          ignore (Unix.umask umask);
+          umask lxor 0o666
+    in
+    if perm <> 0o600 then Unix.chmod tmp_name perm;
+    Sys.rename tmp_name file
+  with
+    e ->
       close_out_noerr oc;
-      begin try Sys.remove tmp_name with _ -> () end;
       raise e
-  end;
-  Sys.rename tmp_name file
 
 let output_sexp_nl do_output oc sexp =
   do_output oc sexp;
@@ -274,14 +241,12 @@ let to_string = to_string_mach
 (* Scan functions *)
 
 let scan_sexp ?buf lexbuf = Parser.sexp (Lexer.main ?buf) lexbuf
-let scan_sexp_opt ?buf lexbuf = Parser.sexp_opt (Lexer.main ?buf) lexbuf
 let scan_sexps ?buf lexbuf = Parser.sexps (Lexer.main ?buf) lexbuf
-let scan_rev_sexps ?buf lexbuf = Parser.rev_sexps (Lexer.main ?buf) lexbuf
 
 let get_main_buf buf =
   let buf =
     match buf with
-    | None -> Buffer.create 128
+    | None -> Buffer.create 64
     | Some buf -> buf in
   Lexer.main ~buf
 
@@ -351,30 +316,20 @@ module Parse_pos = struct
     else if text_char < 0 then fail "text_char < 0"
     else if global_offset < 0 then fail "global_offset < 0"
     else if buf_pos < 0 then fail "buf_pos < 0"
-    else { text_line; text_char; global_offset; buf_pos }
+    else
+      {
+        text_line = text_line;
+        text_char = text_char;
+        global_offset = global_offset;
+        buf_pos = buf_pos;
+      }
 
-  let with_buf_pos t buf_pos = { t with buf_pos }
-end
-
-module Cont_state = struct
-  type t =
-    | Parsing_whitespace
-    | Parsing_atom
-    | Parsing_list
-    | Parsing_sexp_comment
-    | Parsing_block_comment
-
-  let to_string = function
-    | Parsing_whitespace -> "Parsing_whitespace"
-    | Parsing_atom -> "Parsing_atom"
-    | Parsing_list -> "Parsing_list"
-    | Parsing_sexp_comment -> "Parsing_sexp_comment"
-    | Parsing_block_comment -> "Parsing_block_comment"
+  let with_buf_pos t buf_pos = { t with buf_pos = buf_pos }
 end
 
 type ('a, 't) parse_result =
   | Done of 't * Parse_pos.t
-  | Cont of Cont_state.t * ('a, 't) parse_fun
+  | Cont of bool * ('a, 't) parse_fun
 
 and ('a, 't) parse_fun = pos : int -> len : int -> 'a -> ('a, 't) parse_result
 
@@ -398,11 +353,11 @@ type parse_error =
 
 exception Parse_error of parse_error
 
-let bump_text_line { parse_pos; _ } =
+let bump_text_line { parse_pos = parse_pos } =
   parse_pos.Parse_pos.text_line <- parse_pos.Parse_pos.text_line + 1;
   parse_pos.Parse_pos.text_char <- 0
 
-let bump_text_pos { parse_pos; _ } =
+let bump_text_pos { parse_pos = parse_pos } =
   parse_pos.Parse_pos.text_char <- parse_pos.Parse_pos.text_char + 1
 
 let bump_pos_cont state str ~max_pos ~pos cont =
@@ -429,25 +384,34 @@ let set_parse_pos parse_pos buf_pos =
   parse_pos.Parse_pos.buf_pos <- buf_pos;
   parse_pos.Parse_pos.global_offset <- parse_pos.Parse_pos.global_offset + len
 
-let mk_parse_pos { parse_pos; _ } buf_pos =
+let mk_parse_pos { parse_pos = parse_pos } buf_pos =
   set_parse_pos parse_pos buf_pos;
   parse_pos
 
 let raise_parse_error parse_state location buf_pos err_msg =
-  match parse_state with
-  | `Sexp { parse_pos; _ } | `Annot { parse_pos; _ } ->
-      set_parse_pos parse_pos buf_pos;
-      let parse_error = { location; err_msg; parse_state } in
-      raise (Parse_error parse_error)
+  begin
+    match parse_state with
+    | `Sexp { parse_pos = parse_pos }
+    | `Annot { parse_pos = parse_pos } ->
+        set_parse_pos parse_pos buf_pos;
+        parse_pos.Parse_pos.text_char <- parse_pos.Parse_pos.text_char + 1;
+  end;
+  let parse_error =
+    {
+      location = location;
+      err_msg = err_msg;
+      parse_state = parse_state;
+    }
+  in
+  raise (Parse_error parse_error)
 
 let raise_unexpected_char parse_state location buf_pos c =
   let err_msg = sprintf "unexpected character: '%c'" c in
   raise_parse_error parse_state location buf_pos err_msg
 
-let mk_cont_parser cont_parse = (); fun _state str ~max_pos ~pos ->
-  let len = max_pos - pos + 1 in
-  cont_parse ~pos ~len str
-
+(* The code below is derived from lexer.mll in the OCaml distribution,
+   which also uses ASCII codes instead of escape sequences to denote
+   special characters. *)
 
 (* Macro for generating parsers *)
 #define MK_PARSER( \
@@ -476,7 +440,8 @@ let mk_cont_parser cont_parse = (); fun _state str ~max_pos ~pos ->
     if pos_len > str_len then invalid_arg (loc ^ ": pos + len > str_len"); \
     pos_len - 1 \
   \
-  let mk_cont_state name cont state ~cont_state = \
+  let mk_cont name cont state = \
+    let ws_only = GET_PSTACK = [] && Buffer.length state.pbuf = 0 in \
     let parse_fun = \
       let used_ref = ref false in \
       fun ~pos ~len str -> \
@@ -488,16 +453,7 @@ let mk_cont_parser cont_parse = (); fun _state str ~max_pos ~pos ->
           cont state str ~max_pos ~pos \
         end \
     in \
-    Cont (cont_state, parse_fun) \
-  \
-  let mk_cont name cont state = \
-    let cont_state = \
-      match GET_PSTACK = [], Buffer.length state.pbuf = 0 with \
-      | true, true -> Cont_state.Parsing_whitespace \
-      | false, true -> Cont_state.Parsing_list \
-      | _, false -> Cont_state.Parsing_atom \
-    in \
-    mk_cont_state name cont state ~cont_state \
+    Cont (ws_only, parse_fun) \
   \
   let rec PARSE state str ~max_pos ~pos = \
     if pos > max_pos then mk_cont "parse" PARSE state \
@@ -524,146 +480,35 @@ let mk_cont_parser cont_parse = (); fun _state str ~max_pos ~pos ->
                   bump_pos_cont state str ~max_pos ~pos PARSE) \
       | ' ' | '\009' | '\012' -> bump_pos_cont state str ~max_pos ~pos PARSE \
       | '\010' -> bump_line_cont state str ~max_pos ~pos PARSE \
-      | '\013' -> bump_pos_cont state str ~max_pos ~pos parse_nl \
+      | '\013' -> bump_line_cont state str ~max_pos ~pos parse_nl \
       | ';' -> bump_pos_cont state str ~max_pos ~pos parse_comment \
       | '"' -> \
           REGISTER_POS1 \
           bump_pos_cont state str ~max_pos ~pos parse_quoted \
       | c -> \
           REGISTER_POS \
-          let parse = \
-            match c with \
-            | '#' -> maybe_parse_comment \
-            | '|' -> maybe_parse_close_comment \
-            | _ -> parse_atom \
-          in \
-          add_bump_pos state str ~max_pos ~pos c parse \
+          add_bump_pos state str ~max_pos ~pos c parse_atom \
   \
   and parse_nl state str ~max_pos ~pos = \
     if pos > max_pos then mk_cont "parse_nl" parse_nl state \
     else \
-      let c = GET_CHAR in \
-      if c = '\010' then bump_line_cont state str ~max_pos ~pos PARSE \
-      else raise_unexpected_char (MK_PARSE_STATE state) "parse_nl" pos c \
+      let pos = if GET_CHAR = '\010' then pos + 1 else pos in \
+      PARSE state str ~max_pos ~pos \
   \
   and parse_comment state str ~max_pos ~pos = \
     if pos > max_pos then mk_cont "parse_comment" parse_comment state \
     else \
       match GET_CHAR with \
       | '\010' -> bump_line_cont state str ~max_pos ~pos PARSE \
-      | '\013' -> bump_pos_cont state str ~max_pos ~pos parse_nl \
+      | '\013' -> bump_line_cont state str ~max_pos ~pos parse_nl \
       | _ -> bump_pos_cont state str ~max_pos ~pos parse_comment \
   \
-  and maybe_parse_comment state str ~max_pos ~pos = \
-    if pos > max_pos then \
-      mk_cont "maybe_parse_comment" maybe_parse_comment state \
-    else \
-      match GET_CHAR with \
-      | ';' -> bump_pos_cont state str ~max_pos ~pos parse_sexp_comment \
-      | '|' -> bump_pos_cont state str ~max_pos ~pos parse_block_comment \
-      | _ -> parse_atom state str ~max_pos ~pos \
-  \
-  and maybe_parse_close_comment state str ~max_pos ~pos = \
-    if pos > max_pos then \
-      mk_cont "maybe_parse_close_comment" maybe_parse_close_comment state \
-    else \
-      if GET_CHAR <> '#' then parse_atom state str ~max_pos ~pos \
-      else \
-        let err_msg = "end of block comment without start" in \
-        raise_parse_error (MK_PARSE_STATE state) \
-          "maybe_parse_close_comment" pos err_msg \
-  \
-  and parse_sexp_comment state str ~max_pos ~pos = \
-    let pbuf_str = "" in \
-    ignore (MK_ATOM); \
-    Buffer.clear state.pbuf; \
-    let old_pstack = GET_PSTACK in \
-    let pstack = [] in \
-    SET_PSTACK; \
-    let rec loop parse state str ~max_pos ~pos = \
-      match parse state str ~max_pos ~pos with \
-      | Done (_sexp, { Parse_pos.buf_pos = pos; _ }) -> \
-          Buffer.clear state.pbuf; \
-          let pstack = old_pstack in \
-          SET_PSTACK; \
-          PARSE state str ~max_pos ~pos \
-      | Cont (_, cont_parse) -> \
-          Buffer.clear state.pbuf; \
-          let parse = mk_cont_parser cont_parse in \
-          mk_cont_state "parse_sexp_comment" (loop parse) state \
-            ~cont_state:Cont_state.Parsing_sexp_comment \
-    in \
-    loop PARSE state str ~max_pos ~pos \
-  \
-  and parse_block_comment state str ~max_pos ~pos = \
-    let pbuf_str = "" in \
-    ignore (MK_ATOM); \
-    Buffer.clear state.pbuf; \
-    let old_pstack = GET_PSTACK in \
-    let pstack = [] in \
-    SET_PSTACK; \
-    let rec loop depth state str ~max_pos ~pos = \
-      let rec parse_block_depth state str ~max_pos ~pos = \
-        if pos > max_pos then \
-          mk_cont "parse_block_depth" parse_block_depth state \
-        else \
-          match GET_CHAR with \
-          | '\010' -> bump_line_cont state str ~max_pos ~pos parse_block_depth \
-          | '"' -> \
-              REGISTER_POS1 \
-              let rec parse_block_quote parse state str ~max_pos ~pos = \
-                match parse state str ~max_pos ~pos with \
-                | Done (_sexp, { Parse_pos.buf_pos = pos; _ }) -> \
-                    Buffer.clear state.pbuf; \
-                    parse_block_depth state str ~max_pos ~pos \
-                | Cont (_, cont_parse) -> \
-                    Buffer.clear state.pbuf; \
-                    let parse = mk_cont_parser cont_parse in \
-                    mk_cont_state "parse_block_quote" \
-                      (parse_block_quote parse) state \
-                      ~cont_state:Cont_state.Parsing_block_comment \
-              in \
-              bump_pos_cont state str ~max_pos ~pos \
-                (parse_block_quote parse_quoted) \
-          | '#' -> bump_pos_cont state str ~max_pos ~pos parse_open_block \
-          | '|' -> bump_pos_cont state str ~max_pos ~pos parse_close_block \
-          | _ -> bump_pos_cont state str ~max_pos ~pos parse_block_depth \
-      and parse_open_block state str ~max_pos ~pos = \
-        if pos > max_pos then \
-          mk_cont "parse_open_block" parse_open_block state \
-        else \
-          if GET_CHAR = '|' then \
-            bump_pos_cont state str ~max_pos ~pos (loop (depth + 1)) \
-          else parse_block_depth state str ~max_pos ~pos \
-      and parse_close_block state str ~max_pos ~pos = \
-        if pos > max_pos then \
-          mk_cont "parse_close_block" parse_close_block state \
-        else if GET_CHAR = '#' then \
-          let parse = \
-            if depth = 1 then \
-              let () = Buffer.clear state.pbuf in \
-              let pstack = old_pstack in \
-              SET_PSTACK; \
-              PARSE \
-            else loop (depth - 1) \
-          in \
-          bump_pos_cont state str ~max_pos ~pos parse \
-        else parse_block_depth state str ~max_pos ~pos \
-      in \
-      parse_block_depth state str ~max_pos ~pos \
-    in \
-    loop 1 state str ~max_pos ~pos \
-  \
   and parse_atom state str ~max_pos ~pos = \
     if pos > max_pos then mk_cont "parse_atom" parse_atom state \
     else \
       match GET_CHAR with \
       | ' ' | '\009' | '\012' -> \
           bump_found_atom bump_text_pos state str ~max_pos ~pos PARSE \
-      | '#' as c -> \
-          add_bump_pos state str ~max_pos ~pos c maybe_parse_bad_atom_hash \
-      | '|' as c -> \
-          add_bump_pos state str ~max_pos ~pos c maybe_parse_bad_atom_pipe \
       | '(' -> \
           let pbuf = state.pbuf in \
           let pbuf_str = Buffer.contents pbuf in \
@@ -696,7 +541,7 @@ let mk_cont_parser cont_parse = (); fun _state str ~max_pos ~pos ->
                   bump_pos_cont state str ~max_pos ~pos PARSE) \
       | '\010' -> bump_found_atom bump_text_line state str ~max_pos ~pos PARSE \
       | '\013' -> \
-          bump_found_atom bump_text_pos state str ~max_pos ~pos parse_nl \
+          bump_found_atom bump_text_line state str ~max_pos ~pos parse_nl \
       | ';' -> \
           bump_found_atom bump_text_pos state str ~max_pos ~pos parse_comment \
       | '"' -> \
@@ -704,32 +549,9 @@ let mk_cont_parser cont_parse = (); fun _state str ~max_pos ~pos ->
             bump_text_pos state str ~max_pos ~pos reg_parse_quoted \
       | c -> add_bump_pos state str ~max_pos ~pos c parse_atom \
   \
-  and maybe_parse_bad_atom_pipe state str ~max_pos ~pos = \
-    if pos > max_pos then \
-      mk_cont "maybe_parse_bad_atom_pipe" maybe_parse_bad_atom_pipe state \
-    else \
-      match GET_CHAR with \
-      | '#' -> \
-          let err_msg = "illegal end of block comment in unquoted atom" in \
-          raise_parse_error (MK_PARSE_STATE state) "maybe_parse_bad_atom_pipe" \
-            pos err_msg \
-      | _ -> parse_atom state str ~max_pos ~pos \
-  \
-  and maybe_parse_bad_atom_hash state str ~max_pos ~pos = \
-    if pos > max_pos then \
-      mk_cont "maybe_parse_bad_atom_hash" maybe_parse_bad_atom_hash state \
-    else \
-      match GET_CHAR with \
-      | '|' -> \
-          let err_msg = "illegal start of block comment in unquoted atom" in \
-          raise_parse_error (MK_PARSE_STATE state) "maybe_parse_bad_atom_hash" \
-            pos err_msg \
-      | _ -> parse_atom state str ~max_pos ~pos \
-  \
   and reg_parse_quoted state str ~max_pos ~pos = \
     REGISTER_POS \
     parse_quoted state str ~max_pos ~pos \
-  \
   and parse_quoted state str ~max_pos ~pos = \
     if pos > max_pos then mk_cont "parse_quoted" parse_quoted state \
     else \
@@ -747,14 +569,28 @@ let mk_cont_parser cont_parse = (); fun _state str ~max_pos ~pos ->
               bump_pos_cont state str ~max_pos ~pos PARSE) \
       | '\\' -> bump_pos_cont state str ~max_pos ~pos parse_escaped \
       | '\010' as c -> add_bump_line state str ~max_pos ~pos c parse_quoted \
+      | '\013' as c -> add_bump_line state str ~max_pos ~pos c parse_quoted_nl \
       | c -> add_bump_pos state str ~max_pos ~pos c parse_quoted \
   \
+  and parse_quoted_nl state str ~max_pos ~pos = \
+    if pos > max_pos then mk_cont "parse_quoted_nl" parse_quoted_nl state \
+    else \
+      let pos = \
+        let c = '\010' in \
+        if GET_CHAR = c then ( \
+          Buffer.add_char state.pbuf c; \
+          pos + 1 \
+        ) \
+        else pos \
+      in \
+      parse_quoted state str ~max_pos ~pos \
+  \
   and parse_escaped state str ~max_pos ~pos = \
     if pos > max_pos then mk_cont "parse_escaped" parse_escaped state \
     else \
       match GET_CHAR with \
       | '\010' -> bump_line_cont state str ~max_pos ~pos parse_skip_ws \
-      | '\013' -> bump_pos_cont state str ~max_pos ~pos parse_skip_ws_nl \
+      | '\013' -> bump_line_cont state str ~max_pos ~pos parse_skip_ws_nl \
       | '0' .. '9' as c -> \
           bump_text_pos state; \
           let d = Char.code c - 48 in \
@@ -782,12 +618,8 @@ let mk_cont_parser cont_parse = (); fun _state str ~max_pos ~pos ->
   and parse_skip_ws_nl state str ~max_pos ~pos = \
     if pos > max_pos then mk_cont "parse_skip_ws_nl" parse_skip_ws_nl state \
     else \
-      if GET_CHAR = '\010' then \
-        bump_line_cont state str ~max_pos ~pos parse_skip_ws \
-      else begin \
-        Buffer.add_char state.pbuf '\013'; \
-        parse_quoted state str ~max_pos ~pos \
-      end \
+      let pos = if GET_CHAR = '\010' then pos + 1 else pos in \
+      parse_skip_ws state str ~max_pos ~pos \
   \
   and parse_dec state str ~max_pos ~pos ~count ~d = \
     if pos > max_pos then mk_cont "parse_dec" (parse_dec ~count ~d) state \
@@ -838,7 +670,7 @@ let mk_cont_parser cont_parse = (); fun _state str ~max_pos ~pos ->
     let max_pos = check_str_bounds "parse" ~pos ~len str in \
     let state = \
       { \
-        parse_pos; \
+        parse_pos = parse_pos; \
         pstack = INIT_PSTACK; \
         pbuf = Buffer.create 128; \
       } \
@@ -861,21 +693,24 @@ let get_glob_ofs parse_pos pos =
   parse_pos.Parse_pos.global_offset + pos - parse_pos.Parse_pos.buf_pos
 
 let mk_annot_pos
-      ({ Parse_pos.text_line = line; text_char = col; _ } as parse_pos) pos =
-  { Annot.line; col; offset = get_glob_ofs parse_pos pos }
+      ({ Parse_pos.text_line = line; text_char = col } as parse_pos) pos =
+  let offset = get_glob_ofs parse_pos pos in
+  { Annot.line = line; col = col; offset = offset }
 
 let mk_annot_pos1
-      ({ Parse_pos.text_line = line; text_char = col; _ } as parse_pos) pos =
-  { Annot.line; col = col + 1; offset = get_glob_ofs parse_pos pos }
+      ({ Parse_pos.text_line = line; text_char = col } as parse_pos) pos =
+  let offset = get_glob_ofs parse_pos pos in
+  { Annot.line = line; col = col + 1; offset = offset }
 
-let add_annot_pos { parse_pos; pstack; pbuf = _ } pos =
-  pstack.Annot.positions <- mk_annot_pos parse_pos pos :: pstack.Annot.positions
+let add_annot_pos { parse_pos = parse_pos; pstack = pstack } pos =
+  pstack.Annot.positions <-
+    mk_annot_pos parse_pos pos :: pstack.Annot.positions
 
-let add_annot_pos1 { parse_pos; pstack; pbuf = _ } pos =
+let add_annot_pos1 { parse_pos = parse_pos; pstack = pstack } pos =
   pstack.Annot.positions <-
     mk_annot_pos1 parse_pos pos :: pstack.Annot.positions
 
-let get_annot_range { parse_pos; pstack; pbuf = _ } pos =
+let get_annot_range { parse_pos = parse_pos; pstack = pstack } pos =
   let start_pos =
     match pstack.Annot.positions with
     | [] -> assert false  (* impossible *)
@@ -889,7 +724,7 @@ let get_annot_range { parse_pos; pstack; pbuf = _ } pos =
       offset = get_glob_ofs parse_pos pos;
     }
   in
-  { Annot.start_pos; end_pos }
+  { Annot.start_pos = start_pos; end_pos = end_pos }
 
 let mk_annot_atom parse_state str pos =
   Annot.Atom (get_annot_range parse_state pos, Atom str)
@@ -938,7 +773,7 @@ let mk_this_parse ?parse_pos my_parse = (); fun ~pos ~len str ->
   let parse_pos =
     match parse_pos with
     | None -> Parse_pos.create ~buf_pos:pos ()
-    | Some parse_pos -> parse_pos.Parse_pos.buf_pos <- pos; parse_pos
+    | Some parse_pos -> Parse_pos.with_buf_pos parse_pos pos
   in
   my_parse ?parse_pos:(Some parse_pos) ?len:(Some len) str
 
@@ -958,30 +793,30 @@ let input_sexp ?parse_pos ic = gen_input_sexp parse ?parse_pos ic
 let gen_input_rev_sexps my_parse ?parse_pos ?(buf = String.create 8192) ic =
   let rev_sexps_ref = ref [] in
   let buf_len = String.length buf in
-  let rec loop this_parse ~pos ~len ~cont_state =
+  let rec loop this_parse ~pos ~len ~is_incomplete =
     if len > 0 then
       match this_parse ~pos ~len buf with
-      | Done (sexp, ({ Parse_pos.buf_pos; _ } as parse_pos)) ->
+      | Done (sexp, ({ Parse_pos.buf_pos = buf_pos } as parse_pos)) ->
           rev_sexps_ref := sexp :: !rev_sexps_ref;
           let n_parsed = buf_pos - pos in
           let this_parse = mk_this_parse ~parse_pos my_parse in
-          let cont_state = Cont_state.Parsing_whitespace in
           if n_parsed = len then
             let new_len = input ic buf 0 buf_len in
-            loop this_parse ~pos:0 ~len:new_len ~cont_state
-          else loop this_parse ~pos:buf_pos ~len:(len - n_parsed) ~cont_state
-      | Cont (cont_state, this_parse) ->
-          loop this_parse ~pos:0 ~len:(input ic buf 0 buf_len) ~cont_state
-    else
-      if cont_state = Cont_state.Parsing_whitespace then !rev_sexps_ref
-      else
-        failwith (
-          "Sexplib.Sexp.input_rev_sexps: reached EOF while in state "
-          ^ Cont_state.to_string cont_state)
+            loop this_parse ~pos:0 ~len:new_len ~is_incomplete:false
+          else
+            loop this_parse
+              ~pos:buf_pos ~len:(len - n_parsed) ~is_incomplete:false
+      | Cont (ws_only, this_parse) ->
+          loop this_parse
+            ~pos:0 ~len:(input ic buf 0 buf_len) ~is_incomplete:(not ws_only)
+    else if is_incomplete then
+      failwith
+        "Sexplib.Sexp.input_rev_sexps: reached EOF with incomplete S-expression"
+    else !rev_sexps_ref
   in
   let len = input ic buf 0 buf_len in
   let this_parse = mk_this_parse ?parse_pos my_parse in
-  loop this_parse ~pos:0 ~len ~cont_state:Cont_state.Parsing_whitespace
+  loop this_parse ~pos:0 ~len ~is_incomplete:false
 
 let input_rev_sexps ?parse_pos ?buf ic =
   gen_input_rev_sexps parse ?parse_pos ?buf ic
@@ -994,7 +829,7 @@ let input_sexps ?parse_pos ?buf ic =
 
 let of_string_bigstring loc this_parse ws_buf get_len get_sub str =
   match this_parse str with
-  | Done (_, { Parse_pos.buf_pos; _ }) when buf_pos <> get_len str ->
+  | Done (_, { Parse_pos.buf_pos = buf_pos }) when buf_pos <> get_len str ->
       let prefix_len = min (get_len str - buf_pos) 20 in
       let prefix = get_sub str buf_pos prefix_len in
       let msg =
@@ -1004,18 +839,17 @@ let of_string_bigstring loc this_parse ws_buf get_len get_sub str =
       in
       failwith msg
   | Done (sexp, _) -> sexp
-  | Cont (_, this_parse) ->
+  | Cont (ws_only, this_parse) ->
+      if ws_only then failwith (sprintf "Sexplib.Sexp.%s: whitespace only" loc);
       (* When parsing atoms, the incremental parser cannot tell whether
-         it is at the end until it hits whitespace.  We therefore feed it
-         one space to determine whether it is finished. *)
+         it is at the end until it hits whitespace.  We therefore feed
+         it one space to determine whether it is finished. *)
       match this_parse ~pos:0 ~len:1 ws_buf with
       | Done (sexp, _) -> sexp
-      | Cont (cont_state, _) ->
-          let cont_state_str = Cont_state.to_string cont_state in
+      | Cont _ ->
           failwith (
-            sprintf
-              "Sexplib.Sexp.%s: incomplete S-expression while in state %s: %s"
-              loc cont_state_str (get_sub str 0 (get_len str)))
+            sprintf "Sexplib.Sexp.%s: got incomplete S-expression: %s"
+              loc (get_sub str 0 (get_len str)))
 
 let of_string str =
   of_string_bigstring "of_string" parse " " String.length String.sub str
@@ -1047,45 +881,36 @@ let load_rev_sexps ?buf file = gen_load_rev_sexps input_rev_sexps ?buf file
 
 let load_sexps ?buf file = List.rev (load_rev_sexps ?buf file)
 
-let gen_load_sexp_loc = "Sexplib.Sexp.gen_load_sexp"
-
 let gen_load_sexp my_parse ?(strict = true) ?(buf = String.create 8192) file =
   let buf_len = String.length buf in
   let ic = open_in file in
-  let rec loop this_parse ~cont_state =
+  let rec loop this_parse =
     let len = input ic buf 0 buf_len in
     if len = 0 then
-      failwith (
-        sprintf "%s: EOF in %s while in state %s"
-          gen_load_sexp_loc file (Cont_state.to_string cont_state))
+      failwith (sprintf "Sexplib.Sexp.gen_load_sexp: end of file: %s" file)
     else
       match this_parse ~pos:0 ~len buf with
-      | Done (sexp, ({ Parse_pos.buf_pos; _ } as parse_pos)) when strict ->
+      | Done (sexp, ({ Parse_pos.buf_pos = buf_pos } as parse_pos))
+        when strict ->
           let rec strict_loop this_parse ~pos ~len =
             match this_parse ~pos ~len buf with
-            | Done _ ->
+            | Done _ | Cont (false, _) ->
                 failwith (
-                  sprintf "%s: more than one S-expression in file %s"
-                    gen_load_sexp_loc file)
-            | Cont (cont_state, this_parse) ->
+                  sprintf
+                    "Sexplib.Sexp.gen_load_sexp: more than one S-expression: %s"
+                      file)
+            | Cont (true, this_parse) ->
                 let len = input ic buf 0 buf_len in
-                if len > 0 then strict_loop this_parse ~pos:0 ~len
-                else if cont_state = Cont_state.Parsing_whitespace then sexp
-                else
-                  failwith (
-                    sprintf "%s: %s in state %s loading file %s"
-                      gen_load_sexp_loc "additional incomplete data"
-                      (Cont_state.to_string cont_state) file)
+                if len = 0 then sexp
+                else strict_loop this_parse ~pos:0 ~len
           in
           let this_parse = mk_this_parse ~parse_pos my_parse in
           strict_loop this_parse ~pos:buf_pos ~len:(len - buf_pos)
       | Done (sexp, _) -> sexp
-      | Cont (cont_state, this_parse) -> loop this_parse ~cont_state
+      | Cont (_, this_parse) -> loop this_parse
   in
   try
-    let sexp =
-      loop (mk_this_parse my_parse) ~cont_state:Cont_state.Parsing_whitespace
-    in
+    let sexp = loop (mk_this_parse my_parse) in
     close_in ic;
     sexp
   with exc -> close_in_noerr ic; raise exc
@@ -1129,8 +954,9 @@ module Annotated = struct
 
   let get_conv_exn ~file ~exc annot_sexp =
     let range = get_range annot_sexp in
-    let { start_pos = { line; col; offset = _ }; end_pos = _ } = range in
-    let loc = sprintf "%s:%d:%d" file line col in
+    let { start_pos = start_pos } = range in
+    let { line = sline; col = scol } = start_pos in
+    let loc = sprintf "%s:%d:%d" file sline scol in
     Of_sexp_error (Annot.Conv_exn (loc, exc), get_sexp annot_sexp)
 end
 
@@ -1151,27 +977,26 @@ let load_sexp_conv_exn ?strict ?buf file f =
 let load_sexps_conv ?(buf = String.create 8192) file f =
   let rev_sexps = load_rev_sexps ~buf file in
   try List.rev_map (fun sexp -> `Result (f sexp)) rev_sexps
-  with Of_sexp_error _ as e ->
-    match Annotated.load_rev_sexps ~buf file with
-    | [] ->
-        (* File is now empty - perhaps it was a temporary file handle? *)
-        raise e
-    | rev_annot_sexps ->
-        List.rev_map (fun annot_sexp -> Annotated.conv f annot_sexp)
-          rev_annot_sexps
+  with Of_sexp_error (e, s) ->
+    let rev_annot_sexps = Annotated.load_rev_sexps ~buf file in
+    if 0 = List.length rev_annot_sexps then
+      (* File is now empty - perhaps it was a temporary file handle? *)
+      raise (Of_sexp_error (e, s))
+    else
+      List.rev_map (fun annot_sexp -> Annotated.conv f annot_sexp) rev_annot_sexps
 
 let load_sexps_conv_exn ?(buf = String.create 8192) file f =
   let rev_sexps = load_rev_sexps ~buf file in
   try List.rev_map f rev_sexps
-  with Of_sexp_error _ as e ->
-    match Annotated.load_rev_sexps ~buf file with
-    | [] ->
-        (* File is now empty - perhaps it was a temporary file handle? *)
-        raise e
-    | rev_annot_sexps ->
-        List.rev_map
-          (fun annot_sexp -> raise_conv_exn ~file (Annotated.conv f annot_sexp))
-          rev_annot_sexps
+  with Of_sexp_error (e,s) ->
+    let rev_annot_sexps = Annotated.load_rev_sexps ~buf file in
+    if List.length rev_annot_sexps = 0 then
+      (* File is now empty - perhaps it was a temporary file handle? *)
+      raise (Of_sexp_error (e,s))
+    else
+      List.rev_map
+        (fun annot_sexp -> raise_conv_exn ~file (Annotated.conv f annot_sexp))
+        rev_annot_sexps
 
 let gen_of_string_conv of_string annot_of_string str f =
   let sexp = of_string str in
@@ -1194,7 +1019,12 @@ let gen_of_string_conv_exn of_string str f =
   let sexp = of_string str in
   try f sexp
   with Of_sexp_error (exc, sub_sexp) ->
-    raise (Of_string_conv_exn.E { Of_string_conv_exn.exc; sexp; sub_sexp })
+    raise (
+      Of_string_conv_exn.E {
+        Of_string_conv_exn.exc = exc;
+        sexp = sexp;
+        sub_sexp = sub_sexp;
+      })
 
 let of_string_conv_exn str f = gen_of_string_conv_exn of_string str f
 let of_bigstring_conv_exn bstr f = gen_of_string_conv_exn of_bigstring bstr f
@@ -1234,15 +1064,16 @@ let rec subst_found sexp ~subst = function
   | `Pos (pos, found) ->
       match sexp with
       | Atom _ ->
-          failwith "Sexplib.Sexp.subst_found: atom when position requested"
+          failwith
+            "Sexplib.Sexp.subst_search_result: atom when position requested"
       | List lst ->
           let rec loop acc pos = function
             | [] ->
                 failwith
-                  "Sexplib.Sexp.subst_found: short list when position requested"
+                  "Sexplib.Sexp.subst_search_result: \
+                  short list when position requested"
             | h :: t when pos <> 0 -> loop (h :: acc) (pos - 1) t
             | h :: t ->
                 List (List.rev_append acc (subst_found h ~subst found :: t))
           in
           loop [] pos lst
-
